/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/connection.h"
#include "caosdb/configuration.h"                 // for ConnectionConfigur...
#include "caosdb/exceptions.h"                    // for ConfigurationError
#include "caosdb/info.h"                          // for VersionInfo
#include "caosdb/info/v1alpha1/main.grpc.pb.h"    // for GeneralInfoService
#include "caosdb/info/v1alpha1/main.pb.h"         // for GetVersionInfoRequest
#include "caosdb/transaction.h"                   // for Transaction
#include "caosdb/transaction_status.h"            // for TransactionStatus
#include "grpcpp/impl/codegen/status_code_enum.h" // for StatusCode, UNAUTH...
#include <grpcpp/create_channel.h>                // for CreateChannel
#include <grpcpp/impl/codegen/client_context.h>   // for ClientContext
#include <grpcpp/impl/codegen/status.h>           // for Status
#include <string>                                 // for string, operator+

namespace caosdb::connection {
using caosdb::configuration::ConfigurationManager;
using caosdb::configuration::ConnectionConfiguration;
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::entity::v1alpha1::FileTransmissionService;
using caosdb::info::VersionInfo;
using caosdb::info::v1alpha1::GeneralInfoService;
using caosdb::info::v1alpha1::GetVersionInfoRequest;
using caosdb::info::v1alpha1::GetVersionInfoResponse;
using caosdb::transaction::Transaction;
using caosdb::transaction::TransactionStatus;

Connection::Connection(const ConnectionConfiguration &configuration) {
  const std::string target =
    configuration.GetHost() + ":" + std::to_string(configuration.GetPort());
  this->channel =
    grpc::CreateChannel(target, configuration.GetChannelCredentials());
  this->general_info_service = GeneralInfoService::NewStub(this->channel);
  this->entity_transaction_service =
    std::make_shared<EntityTransactionService::Stub>(this->channel);
  this->file_transmission_service =
    std::make_shared<FileTransmissionService::Stub>(this->channel);
}

auto Connection::RetrieveVersionInfoNoExceptions() const noexcept
  -> TransactionStatus {

  const GetVersionInfoRequest request;
  GetVersionInfoResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->general_info_service->GetVersionInfo(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status =
        TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(
        std::to_string(grpc_status.error_code()) + " - " + error_message);
    }
  } else {
    this->version_info =
      std::make_unique<VersionInfo>(response.release_version_info());
  }

  return status;
}

auto Connection::RetrieveVersionInfo() const -> const VersionInfo & {

  TransactionStatus status = RetrieveVersionInfoNoExceptions();
  status.ThrowExceptionIfError();
  return *GetVersionInfo();
}

[[nodiscard]] auto Connection::CreateTransaction() const
  -> std::unique_ptr<Transaction> {
  auto entity_service = this->entity_transaction_service;
  auto file_service = this->file_transmission_service;
  return std::make_unique<Transaction>(entity_service, file_service);
}

auto ConnectionManager::mHasConnection(const std::string &name) const -> bool {
  auto it = connections.find(name);
  return it != connections.end();
}

auto ConnectionManager::mGetConnection(const std::string &name) const
  -> const std::shared_ptr<Connection> & {
  if (!HasConnection(name)) {
    try {
      auto connection = ConfigurationManager::GetConnectionConfiguration(name);
      connections[name] = std::make_shared<Connection>(*connection.release());
    } catch (const caosdb::exceptions::ConfigurationError &exc) {
      throw caosdb::exceptions::UnknownConnectionError("No connection named '" +
                                                       name + "' present.");
    }
  }
  return this->connections.at(name);
}

auto ConnectionManager::mGetDefaultConnection() const
  -> const std::shared_ptr<Connection> & {
  if (!HasConnection(default_connection_name)) {
    default_connection_name = ConfigurationManager::GetDefaultConnectionName();
    auto default_connection =
      ConfigurationManager::GetDefaultConnectionConfiguration();
    connections[default_connection_name] =
      std::make_shared<Connection>(*default_connection.release());
  }

  return connections.at(default_connection_name);
}

} // namespace caosdb::connection
